--[[
This code creates an I/O filter that automatically detects if the user
uploads jpeg images and resizes the images according to the users
image preferences. The I/O filter hooks into the "io:open" method. All
other methods are directed directly to the DiskIo without any
filtering.

The code uses an undocumented feature in the Web File Server (WFS in
wfm.lua) that makes ba.create.wfs(...,true) return the WFS directory
function instead of creating a BA directory. This code therefore
includes a BA directory function "uploadService" that mimics part of
the original WFS directory function such as session URL management.

The BA directory function "uploadService" creates a "per user"
(i.e. per session) WFS, thus allowing multiple users to upload images
with different resizing options.
--]]

local args = {...}
local io = args[1] -- The DiskIo
local getThumbImgName = args[2]
local app = args[3]
local fmt,sfind,smatch=string.format,string.find,string.match
local tinsert,tconcat,tremove=table.insert,table.concat,table.remove

require"gd" -- GD image lib
assert(type(ba.create.luaio) == "function") -- Must be compiled with lio.c

-- Resize, but keep image proportions.
local function resizeImg(im,maxw,maxh)
   local im2,diff,xw,xh
   local w,h = im:sizeXY()
   if w > maxw then
      diff = w/maxw
      xh = h/diff
      if xh > maxh then
         diff = h/maxh
      end
   elseif h > maxh then
      diff = h/maxh
   end
   if diff then
      xw = w/diff
      xh = h/diff
   else
      xw = w
      xh = h
   end
   im2 = gd.createTrueColor(xw,xh)
   if im2 then
      gd.copyResampled(im2,im,0,0,0,0,xw,xh,w,h)
      return im2
   end
   return nil
end


local function appendLog(cfg,log)
   local s=cfg.stats
   if not s then return end
   if not s.log then s.log={} end
   tinsert(s.log,log)
end



-- The following code, which is run in a thread auto saves the album
-- and prevents album loss if user closes the browser window.
local function autosave(cfg)
   local env,conn = app.openDb()
   if not cfg.key then
      local q=fmt('SELECT key FROM PhotoAlbumNames WHERE name="%s"',
                  cfg.albumtitle)
      local cur=conn:execute(q)
      if cur then
         cfg.key = cur:fetch()
         cur:close()
      end
      if not cfg.key then
         local q=fmt('INSERT INTO PhotoAlbumNames (name) values("%s")',
                     cfg.albumtitle)
         ok,err=conn:execute(q)
         cfg.key=conn:getlastautoid() 
      end
   end
   if cfg.photos then
      local imgname = fmt("/%s",tremove(cfg.photos)):gsub("//","/")
      local fmts='INSERT INTO PhotoAlbums (album,img,info) values(%u,"%s","")'
      local ok,err=conn:execute(fmt(fmts,cfg.key,imgname:gsub('"','""')))
      if not ok then
         tinsert(cfg.photos,imgname)
         trace("Album auto save failed:", imgname, ",",err)
      end
   end
   app.closeDb(env,conn)
end



local function appendPhoto(cfg,imgname)
   if not cfg.photos then
      cfg.photos={}
   end
   tinsert(cfg.photos,imgname)
   --ba.thread.run(function() autosave(cfg) end)
   autosave(cfg)
end


local function imgScalingThread(cfg)
   while true do
      local ok,err
      local name = tremove(cfg.jobQueue)
      if not name then
         cfg.jobQueue=nil
         return
      end
      local lname = io:realpath(name)
      local im = gd.createFromJpeg(lname)
      if im then
         appendLog(cfg, fmt("Scaling %s\n", name))
         im2=resizeImg(im,cfg.w,cfg.h)
         if im2 then
            local fp
            fp,err = io:open(name, "w")
            if fp then
               ok,err=fp:write(im2:jpegStr(cfg.quality or 80))
               if ok then
                  ok,err=fp:close()
                  cfg.stats.images = cfg.stats.images + 1
                  if ok and cfg.thumbw then
                     local thumb=getThumbImgName(name)
                     appendLog(cfg, fmt("Creating %s\n",thumb))
                     im2=resizeImg(im,cfg.thumbw,cfg.thumbh)
                     if im2 then
                        fp,err = io:open(thumb,"w")
                        if fp then
                           ok,err=fp:write(im2:jpegStr(cfg.quality))
                           if ok then
                              ok,err=fp:close()
                              if ok then
                                 im,im2=nil,nil
                                 collectgarbage()
                                 appendPhoto(cfg,name)
                              end
                           end
                        end
                     end
                  end
               end
            end
         end
      end
      collectgarbage()
      if not ok then
         if not err then err = "GD error" end
         appendLog(cfg, fmt("Scaling %s failed: %s\n",name,err))
         io:remove(name)
         io:remove(getThumbImgName(name))
      end
   end
end

local function startImgScaling(cfg,name)
   if not cfg.jobQueue then
      cfg.jobQueue={name}
      ba.thread.create():run(function() imgScalingThread(cfg) end)
   else
      tinsert(cfg.jobQueue,name)
   end
end


-- Create the I/O filter and the WFS directory function.
-- The user's image preferences are stored as an upvalue in 'cfg'
local function createWfmService(cfg)
   local function open(name, mode)
      local size
      local fp,err
      name=fmt("/%s",name):gsub("//" ,"/")
      if not cfg.stats then return nil,"noaccess" end
      if not cfg.imgNames then cfg.imgNames={} end

      if mode =="w" and not sfind(name,"%.DAV/([^/]+)$") then
         if cfg.imgNames[name] then
            appendLog(cfg, fmt("Ignoring duplicate upload: %s!\n", name))
         else
            fp,err = io:open(name,mode)
            if not fp then return nil, err end
            appendLog(cfg, fmt("Uploading %s\n", name))
            cfg.stats.files = cfg.stats.files + 1
            if sfind(name,"%.[jJ][pP][eE]?[gG]$") then
               size=0
            end
         end
      else
         fp,err = io:open(name,mode)
         if not fp then return nil, err end
      end

      local function read(x) return fp:read(x) end
      local function seek(x) return fp:seek(x) end
      local function flush() return fp:flush() end
      local function write(x)
         if fp then
            if size then size = size +#x end
            return fp:write(x)
         end
         return true
      end
      local function close()
         if fp then
            local ok,err=fp:close()
            if ok and size and size > 1024 then
               cfg.imgNames[name]=true
               startImgScaling(cfg,name)
            end
            fp=nil
            if ok then return true end
            return nil,err
         end
         return true
      end
      return {read=read,write=write,seek=seek,flush=flush,close=close}
   end -- End open
   local function files(path) -- Directory iterator
      local iter,err=io:files(path,true)
      if not iter then return nil,err end
      local n,d,t,s
      local function read() n,d,t,s = iter() return n and true or false end
      local function name() return n end
      local function stat() return {name=n,isdir=d,mtime=t,size=s} end
      return {read=read,name=name,stat=stat}
   end
   local function stat(x) return io:stat(x) end
   local function mkdir(x) return io:mkdir(x) end
   local function rmdir(x) return io:rmdir(x) end
   local function remove(x) return io:remove(x) end
   local function hide(x) return io:hide(x) end
   local function realpath(x) return io:realpath(x) end
   local lio = ba.create.luaio({open=open,files=files,stat=stat,mkdir=mkdir,
                                  rmdir=rmdir,remove=remove,
                                  hide=hide,realpath=realpath})
   local wfs = require"extwfs"(lio,".LOCK",50,50)
   local constraints = {}
   local roles={"cmsmanager"}
   local urls={"/*"}
   local methods={}
   constraints.admin={roles=roles,urls=urls,methods=methods}
   local authorizer = bd.db.user.jdb:authorizer()
   authorizer:set(json.encode(constraints))
   wfs:setauth(bd.auth, authorizer)
   return wfs
end -- End createWfmService

-- WFS directory function
local dir
local function uploadService(_ENV,rel)
   local s=request:session()
   if not s then
      local p
      s,p=ba.session(request,rel,true)
      if s then
         rel=p
      end
   end
   if s and s.cms then
      local cms=s.cms
      if not s.onterminate then
         function s.onterminate()
            if cms.imgUplService then
               cms.imgUplService:unlink()
            end
         end
      end
      if cms.imgUplCfg then
         if not cms.imgUplService then
            cms.imgUplService = createWfmService(cms.imgUplCfg)
            dir:insert(cms.imgUplService) -- baseuri() requires this
         end
         return cms.imgUplService:service(request,rel)
      end
   end
   response:sendredirect"/private/manage/images.lsp"
end
dir=ba.create.dir("imagefs")
dir:setfunc(uploadService)
return dir
